import numpy as np
import pytest

import unyt
import yt
from yt.data_objects.selection_objects.region import YTRegion
from yt.loaders import load_particles
from yt.testing import assert_rel_equal
from yt.funcs import setdefaultattr

## Notes on PR:
## pixelation_routines.pyx: pixelize_sph_kernel_projection
# Discussion point: do we want the hsml margin on the z direction?
# it's consistent with Ray and Region selections, I think, 
# but does tend to 'tack on' stuff compared to the nominal depth

## TODO: yt/data_objects/tests/test_sph_data_objects
## also does some slices and gridding. Mostly it's
## just consistency tests, which should be ok, but
## test_arbitrary_grid (line 255) will probably fail:
## I believe the provided answers are wrong.

## Ray dl calc is a different issue. 
# these are tests for the following SPH data operations:
# - 2d projection (axis-aligned, all 3 axes, periodic T/F,
#                  depth on/off) -> runs, tests work
# - 2d projection (off-axis, with/without north vector, periodic T/F,
#                  depth T/F) -> runs, tests work
# - slice projection: (axis-aligned, all 3 axes, periodic T/F)
# - slice projection: (off-axis, with/without north vector, 
#                      periodic T/F)
# - 3d gridding: (only axis-aligned, all 3 axes, periodic T/F) 


## copied from yt/data_objects/tests/test_rays.py
## version from the issue4781_sph_ray_dl branch
## TODO: copy to testing.py or something if this ends 
## up being useful for different tests.
def _cubicspline_python(x):
    '''
    tested: volume integral is 1.
    '''
    # C is 8/pi
    _c = 8. / np.pi
    x = np.asarray(x)
    kernel = np.zeros(x.shape, dtype=x.dtype)
    half1 = np.where(np.logical_and(x >=0., x <= 0.5))
    kernel[half1] = 1. - 6. * x[half1]**2 * (1. - x[half1])
    half2 = np.where(np.logical_and(x > 0.5, x <= 1.0))
    kernel[half2] = 2. * (1. - x[half2])**3
    return kernel * _c

def _integrate_kernel(kernelfunc, b, hsml):
    pre = 1. / hsml**2
    x = b / hsml
    xmax = np.sqrt(1. - x**2)
    xmin = -1. * xmax
    xe = np.linspace(xmin, xmax, 500) # shape: 500, x.shape
    xc = 0.5 * (xe[:-1, ...] + xe[1:, ...])
    dx = np.diff(xe, axis=0)
    spv = kernelfunc(np.sqrt(xc**2 + x**2))
    integral = np.sum(spv * dx, axis=0)
    return pre * integral

def constantmass(i: int, j: int, k: int) -> float:
    return 1.

## TODO: move to testing.py if this ends up in the test suite
def fake_sph_flexible_grid_ds(
        hsml_factor: float = 1.0,
        nperside: int = 3,
        periodic: bool = True, 
        e1hat: np.ndarray[float] = np.array([1, 0, 0]),
        e2hat: np.ndarray[float] = np.array([0, 1, 0]),
        e3hat: np.ndarray[float] = np.array([0, 0, 1]),
        offsets: np.ndarray[float] = 0.5 * np.ones((3,), dtype=np.float64),
        massgenerator = constantmass,
        unitrho: float = 1.,
        bbox: np.ndarray | None = None, 
        recenter: np.ndarray | None = None,
        ):
    """Returns an in-memory SPH dataset useful for testing

    This dataset should have `nperside`**3 particles with the particles
    arranged uniformly on a 3D grid. 
    The basis vectors `e1hat`, `e2hat`, and `e3hat` define this grid.
    (If these are not normalized to 1 or not orthogonal, the spacing or
    overlap between SPH particles will be affected, but this is 
    allowed.) 
    Along these basis vectors, `offsets` give the zero point.
    All particles will have smoothing regions with a radius of 
    `hsml_factor` * 0.5,
    `massgenerator` is a function assigning a mass to each particle,
    as a function of the e[1-3]hat indices, in order 
    `unitrho` defines the density for a particle with mass 1 ('g'),
    and the standard (uniform) grid hsml_factor.
    `bbox` is the assumed enclosing volume
    `recenter`: if not None, after generating the grid, the positions
    are periodically shifted to move the old center to this positions. 
    Useful for testing periodicity handling. 
    This shift is relative to the halfway positions of the bbox edges.
    """

    npart = nperside**3

    pos = np.empty((npart, 3), dtype=np.float64)
    mass = np.empty((npart,), dtype=np.float64)
    for i in range(0, nperside):
        for j in range(0, nperside):
            for k in range(0, nperside):
                _pos = (offsets[0] + i) * e1hat \
                        + (offsets[1] + j) * e2hat \
                        + (offsets[2] + k) * e3hat
                ind = nperside**2 * i + nperside * j + k
                pos[ind, :] = _pos
                mass[ind] = massgenerator(i, j, k)
    rho = unitrho * mass

    if bbox is None:
        eps = 1e-3
        margin = (1. + eps) * hsml_factor
        bbox = np.array([[np.min(pos[:, 0]) - margin,
                          np.max(pos[:, 0]) + margin],
                         [np.min(pos[:, 1]) - margin,
                          np.max(pos[:, 1]) + margin],
                         [np.min(pos[:, 2]) - margin,
                          np.max(pos[:, 2]) + margin], 
                         ])
         
    if recenter is not None:
        periods = bbox[:, 1] - bbox[:, 0]
        # old center -> new position
        pos += -0.5 * periods[np.newaxis, :] + recenter[np.newaxis, :]
        # wrap coordinates -> all in [0, boxsize) range
        pos %= periods[np.newaxis, :] 
        # shift back to original bbox range
        pos += (bbox[:, 0])[np.newaxis, :]
    if not periodic:
        # remove points outside bbox to avoid errors:
        okinds = np.ones(len(mass), dtype=bool)
        for ax in [0, 1, 2]:
            okinds &= pos[:, ax] < bbox[ax, 1]
            okinds &= pos[:, ax] >= bbox[ax, 0]
        npart = sum(okinds)
    else:
        okinds = slice(None, None, None)

    data = {
        "particle_position_x": (np.copy(pos[okinds, 0]), "cm"),
        "particle_position_y": (np.copy(pos[okinds, 1]), "cm"),
        "particle_position_z": (np.copy(pos[okinds, 2]), "cm"),
        "particle_mass": (mass[okinds], "g"),
        "particle_velocity_x": (np.zeros(npart), "cm/s"),
        "particle_velocity_y": (np.zeros(npart), "cm/s"),
        "particle_velocity_z": (np.zeros(npart), "cm/s"),
        "smoothing_length": (np.ones(npart) * 0.5 * hsml_factor, "cm"),
        "density": (rho[okinds], "g/cm**3"),
    }
    
    ds = load_particles(data=data,
                        bbox=bbox, periodicity=(periodic,) * 3,
                        length_unit=1., mass_unit=1., time_unit=1.,
                        velocity_unit=1.)
    ds.kernel_name = 'cubic'
    return ds

########### Along-axis projections onto a grid ############
# column density maps, not slices

def test_sph_projection_basic1():
    '''
    small, uniform grid: expected values for given dl?
    pixel centers at 0.5, 1., 1.5, 2., 2.5
    particles at 0.5, 1.5, 2.5
    '''
    bbox = np.array([[0., 3.]] * 3)
    ds = fake_sph_flexible_grid_ds(hsml_factor=1.0, nperside=3, 
                                   bbox=bbox)
    #reg = YTRegion(center=np.array([1.5, 1.5, 1.5]),
    #               left_edge=np.copy(bbox[:, 0] + 0.25),
    #               right_edge=np.copy(bbox[:, 1] - 0.25),
    #               ds=ds)
    # works, but no depth control (at least without specific filters)
    proj = ds.proj(("gas", "density"), 2) 
    frb = proj.to_frb(width=(2.5, 'cm'), 
                      resolution=(5, 5),
                      height=(2.5, 'cm'),
                      center=np.array([1.5, 1.5, 1.5]), 
                      periodic=False)
    out = frb.get_image(('gas', 'density'))
    
    expected_out = np.zeros((5, 5), dtype=np.float64)
    dl_1part = _integrate_kernel(_cubicspline_python, 0., 0.5)
    linedens_1part = dl_1part * 1. # unit mass, density
    linedens = 3. * linedens_1part
    expected_out[::2, ::2] = linedens

    assert_rel_equal(expected_out, out.v, 5)
    #return out

def test_sph_projection_basic2():
    '''
    small, uniform grid: expected values for given dl?
    pixel centers at 0.5, 1., 1.5, 2., 2.5
    particles at 0.5, 1.5, 2.5
    but hsml radii are 0.25 -> try non-zero impact parameters,
    other pixels are still zero.
    '''
    bbox = np.array([[0., 3.]] * 3)
    ds = fake_sph_flexible_grid_ds(hsml_factor=0.5, nperside=3, 
                                   bbox=bbox)
    proj = ds.proj(("gas", "density"), 2) 
    frb = proj.to_frb(width=(2.5, 'cm'), 
                      resolution=(5, 5),
                      height=(2.5, 'cm'),
                      center=np.array([1.375, 1.375, 1.5]), 
                      periodic=False)
    out = frb.get_image(('gas', 'density'))
    
    expected_out = np.zeros((5, 5), dtype=np.float64)
    dl_1part = _integrate_kernel(_cubicspline_python, 
                                 np.sqrt(2) * 0.125,
                                 0.25)
    linedens_1part = dl_1part * 1. # unit mass, density
    linedens = 3. * linedens_1part
    expected_out[::2, ::2] = linedens

    #print(expected_out)
    #print(out.v)
    assert_rel_equal(expected_out, out.v, 4)
    #return out

def get_dataset_sphrefine(reflevel: int = 1):
    '''
    constant density particle grid, 
    with increasing particle sampling
    '''
    lenfact = (1./3.)**(reflevel - 1)
    massfact = lenfact**3
    nperside = 3**reflevel

    e1hat = np.array([lenfact, 0, 0])
    e2hat = np.array([0, lenfact, 0])
    e3hat = np.array([0, 0, lenfact])
    hsml_factor = lenfact
    bbox = np.array([[0., 3.]] * 3)
    offsets = np.ones(3, dtype=np.float64) * 0.5 # in units of ehat
    
    def refmass(i: int, j: int, k: int) -> float:
        return massfact
    unitrho = 1. / massfact # want density 1 for decreasing mass

    ds = fake_sph_flexible_grid_ds(hsml_factor=hsml_factor, 
                                   nperside=nperside,
                                   periodic=True, 
                                   e1hat=e1hat,
                                   e2hat=e2hat, 
                                   e3hat=e3hat,
                                   offsets=offsets,
                                   massgenerator=refmass,
                                   unitrho=unitrho,
                                   bbox=bbox, 
                                   )
    return ds 


def getdata_test_gridproj2():
    # initial pixel centers at 0.5, 1., 1.5, 2., 2.5
    # particles at 0.5, 1.5, 2.5
    # refine particle grid, check if pixel values remain the 
    # same in the pixels passing through initial particle centers
    outlist = []
    dss = []
    for rl in range(1, 4):
        ds = get_dataset_sphrefine(reflevel=rl)
        proj = ds.proj(("gas", "density"), 2) 
        frb = proj.to_frb(width=(2.5, 'cm'), 
                          resolution=(5, 5),
                          height=(2.5, 'cm'),
                          center=np.array([1.5, 1.5, 1.5]), 
                          periodic=False)
        out = frb.get_image(('gas', 'density'))
        outlist.append(out)
        dss.append(ds)
    
    return outlist, dss

def test_sph_gridproj_reseffect1():
    '''
    Comparing same pixel centers with higher particle resolution.
    The pixel centers are at x/y coordinates [0.5, 1., 1.5, 2., 2.5]
    at the first level, the spacing halves at each level.
    Checking the pixels at [0.5, 1.5, 2.5],
    which should have the same values at each resolution.
    '''
    imgs, _ = getdata_test_gridproj2()
    ref = imgs[-1]
    for i, img in enumerate(imgs):
        #print(img[::img.shape[0] // 2, ::img.shape[1] // 2])
        #print(ref[::ref.shape[0] // 2, ::ref.shape[1] // 2])
        #print(img)
        #print(ref)
        assert_rel_equal(img[::img.shape[0] // 2, ::img.shape[1] // 2],
                         ref[::ref.shape[0] // 2, ::ref.shape[1] // 2], 4)
    #print('\nComparing same pixel centers with higher particle resolution. '
    #      'The pixel centers are at x/y coordinates [0.5, 1., 1.5, 2., 2.5].'
    #      ' Showing the pixels at [0.5, 1.5, 2.5], '
    #      'which should have the same values at each resolution.')
    #for i, img in enumerate(imgs):
    #    print(f'At refinement level {i + 1}:')
    #    print(img[::2, ::2])
    # print('\n')
    # print('Info on the used datasets:')
    # for i, ds in enumerate(dss):
    #     print(f'At refinement level {i + 1}')
    #     offsets = np.min(ds.r[('gas', 'position')], axis=0)
    #     print('first pixel center in x/y/z: ', offsets)
    #     xposs = np.unique(ds.r[('gas', 'position')][:, 0])
    #     xposs.sort()
    #     xspace = np.unique(np.diff(xposs))
    #     yposs = np.unique(ds.r[('gas', 'position')][:, 1])
    #     yposs.sort()
    #     yspace = np.unique(np.diff(yposs))
    #     zposs = np.unique(ds.r[('gas', 'position')][:, 2])
    #     zposs.sort()
    #     zspace = np.unique(np.diff(zposs))
    #     print('Spacing along the x/y/z axis: ', xspace, yspace, zspace)
    #     hsml = np.unique(ds.r[('gas', 'smoothing_length')])
    #     print('Smoothing lengths: ', hsml)
    #     mass = ds.r[('gas', 'mass')]
    #     print('Particle mass: ', np.unique(mass))
    #     print('Total mass: ', mass.sum())
    #     density = np.unique(ds.r[('gas', 'density')])
    #     print('Particle densities: ', density)
    #     print('\n')


def test_sph_gridproj_reseffect2():
    '''
    refine the pixel grid instead of the particle grid
    '''
    ds = get_dataset_sphrefine(reflevel=2)
    proj = ds.proj(("gas", "density"), 2) 
    imgs = {}
    maxrl = 5
    for rl in range(1, maxrl + 1):
        npix = 1 + 2**(rl + 1)
        margin = 0.5 - 0.5**(rl + 1)
        frb = proj.to_frb(width=(3. - 2. * margin, 'cm'), 
                          resolution=(npix, npix),
                          height=(3. - 2. * margin, 'cm'),
                          center=np.array([1.5, 1.5, 1.5]), 
                          periodic=False)
        out = frb.get_image(('gas', 'density'))
        imgs[rl] = out
    ref = imgs[maxrl]
    pixspace_ref = 2**(maxrl)
    for rl in imgs:
        img = imgs[rl]
        pixspace = 2**(rl)
        #print(f'Grid refinement level {rl}:') 
        assert_rel_equal(img[::pixspace, ::pixspace], 
                         ref[::pixspace_ref, ::pixspace_ref], 4)


@pytest.mark.parametrize("weighted", [True, False])
@pytest.mark.parametrize("periodic", [True, False])
@pytest.mark.parametrize("depth", [None, (1., "cm")])
@pytest.mark.parametrize("shiftcenter", [False, True])
@pytest.mark.parametrize("axis", [0, 1, 2])
def test_sph_proj_general_alongaxes(axis, shiftcenter, depth, periodic, weighted):
    '''
    The previous projection tests were for a specific issue. 
    Here, we test more functionality of the projections.
    We just send lines of sight through pixel centers for convenience.
    Particles at [0.5, 1.5, 2.5] (in each coordinate)
    smoothing lengths 0.25
    all particles have mass 1., density 1.5, 
    except the single center particle, with mass 2., density 3.
    
    Parameters:
    -----------
    axis: {0, 1, 2}
        projection axis (aligned with sim. axis)
    shiftcenter: bool
        shift the coordinates to center the projection on. 
        (The grid is offset to this same center)
    depth: float or None
        depth of the projection slice
    periodic: bool
        assume periodic boundary conditions, or not
    weighted: bool
        make a weighted projection (density-weighted density), or not
    '''
    if shiftcenter:
        center = unyt.unyt_array(np.array((0.625, 0.625, 0.625)), 'cm')
    else:
        center = unyt.unyt_array(np.array((1.5, 1.5, 1.5)), 'cm')
    bbox = unyt.unyt_array(np.array([[0., 3.], [0., 3.], [0., 3.]]), 'cm')
    hsml_factor = 0.5
    unitrho = 1.5
    # test correct centering, particle selection
    def makemasses(i, j, k):
        if i == j == k == 1:
            return 2. 
        else:
            return 1.
    # m / rho, factor 1. / hsml**2 is included in the kernel integral
    # (density is adjusted, so same for center particle)
    prefactor = 1. / unitrho #/ (0.5 * 0.5)**2
    dl_cen = _integrate_kernel(_cubicspline_python, 0., 0.25)
    #dl_shift = _integrate_kernel(_cubicspline_python, 
    #                             np.sqrt(2) * 0.125, 0.25)

    # result shouldn't depend explicitly on the center if we re-center
    # the data, unless we get cut-offs in the non-periodic case 
    ds = fake_sph_flexible_grid_ds(hsml_factor=hsml_factor, nperside=3,
                                   periodic=periodic, 
                                   offsets=np.full(3, 0.5),
                                   massgenerator=makemasses,
                                   unitrho=unitrho,
                                   bbox=bbox.v, 
                                   recenter=center.v)
    if depth is None:
        source = ds.all_data()
    else:
        depth = unyt.unyt_quantity(*depth)
        le = np.array(ds.domain_left_edge)
        re = np.array(ds.domain_right_edge)
        le[axis] = center[axis] - 0.5 * depth
        re[axis] = center[axis] + 0.5 * depth
        cen = 0.5 * (le + re)
        reg = YTRegion(center=cen,
                       left_edge=le,
                       right_edge=re,
                       ds=ds)
        source = reg

    # we don't actually want a plot, it's just a straightforward, 
    # common way to get an frb / image array
    if weighted: 
        toweight_field = ("gas", "density")
    else:
        toweight_field = None
    prj = yt.ProjectionPlot(ds, axis, ("gas", "density"),
                            width=(2.5, "cm"),
                            weight_field=toweight_field,
                            buff_size=(5, 5),
                            center=center,
                            data_source=source)
    img = prj.frb.data[('gas', 'density')]
    if weighted:
        expected_out = np.zeros((5, 5,), dtype=img.v.dtype)
        expected_out[::2, ::2] = unitrho
        if depth is None: 
            ## during shift, particle coords do wrap around edges
            #if (not periodic) and shiftcenter:
            #    # weight 1. for unitrho, 2. for 2. * untrho
            #    expected_out[2, 2] *= 5. / 3.
            #else:
            # weight (2 * 1.) for unitrho, (1 * 2.) for 2. * unitrho
            expected_out[2, 2] *= 1.5
        else:
            # only 2 * unitrho element included
            expected_out[2, 2] *= 2.
    else:
        expected_out = np.zeros((5, 5,), dtype=img.v.dtype)
        expected_out[::2, ::2] = dl_cen * prefactor * unitrho
        if depth is None:
            ## during shift, particle coords do wrap around edges
            #if (not periodic) and shiftcenter:
            #    # 2 particles per l.o.s., including the denser one
            #    expected_out *= 2.
            #    expected_out[2, 2] *= 1.5
            #else:
            # 3 particles per l.o.s., including the denser one
            expected_out *= 3.
            expected_out[2, 2] *= 4. / 3.
        else:
            # 1 particle per l.o.s., including the denser one
            expected_out[2, 2] *= 2.
    # grid is shifted to the left -> 'missing' stuff at the left
    if (not periodic) and shiftcenter: 
        expected_out[:1, :] = 0.
        expected_out[:, :1] = 0.
    #print(axis, shiftcenter, depth, periodic, weighted)
    #print(expected_out)
    #print(img.v)
    assert_rel_equal(expected_out, img.v, 5)

########### Off-axis projections onto a grid ############
# column density maps, not slices

@pytest.mark.parametrize("weighted", [True, False])
@pytest.mark.parametrize("periodic", [True, False])
@pytest.mark.parametrize("depth", [None, (1., "cm"), (0.5, "cm")])
@pytest.mark.parametrize("shiftcenter", [False, True])
@pytest.mark.parametrize("northvector", [None, (1.e-4, 1., 0.)])
def test_sph_proj_general_offaxis(northvector, shiftcenter, depth, periodic, 
                                  weighted):
    '''
    Same as the on-axis projections, but we rotate the basis vectors
    to test whether roations are handled ok. the rotation is chosen to
    be small so that in/exclusion of particles within bboxes, etc. 
    works out the same way. 
    We just send lines of sight through pixel centers for convenience.
    Particles at [0.5, 1.5, 2.5] (in each coordinate)
    smoothing lengths 0.25
    all particles have mass 1., density 1.5, 
    except the single center particle, with mass 2., density 3.
    
    Parameters:
    -----------
    northvector: tuple
        y-axis direction in the final plot (direction vector)
    shiftcenter: bool
        shift the coordinates to center the projection on. 
        (The grid is offset to this same center)
    depth: float or None
        depth of the projection slice
    periodic: bool
        assume periodic boundary conditions, or not
    weighted: bool
        make a weighted projection (density-weighted density), or not
    '''
    #TODO: actually implement
    if shiftcenter:
        center = np.array((0.625, 0.625, 0.625)) # cm
    else:
        center = np.array((1.5, 1.5, 1.5)) # cm
    bbox = unyt.unyt_array(np.array([[0., 3.], [0., 3.], [0., 3.]]), 'cm')
    hsml_factor = 0.5
    unitrho = 1.5
    # test correct centering, particle selection
    def makemasses(i, j, k):
        if i == j == k == 1:
            return 2. 
        else:
            return 1.
    #dl_shift = _integrate_kernel(_cubicspline_python, 
    #                             np.sqrt(2) * 0.125, 0.25)

    # result shouldn't depend explicitly on the center if we re-center
    # the data, unless we get cut-offs in the non-periodic case 
    # *almost* the z-axis
    # try to make sure dl differences from periodic wrapping are small
    epsilon = 1e-4 
    projaxis = np.array([epsilon, 0.00, np.sqrt(1. - epsilon**2)])
    e1dir = projaxis / np.sqrt(np.sum(projaxis**2))
    # TODO: figure out other (default) axes for basis vectors here
    if northvector is None:
        e2dir = np.array([0., 1., 0.])
    else:
        e2dir = np.asarray(northvector)
    e2dir = e2dir - np.sum(e1dir * e2dir) * e2dir # orthonormalize
    e2dir /= np.sqrt(np.sum(e2dir**2))
    e3dir = np.cross(e1dir, e2dir)

    ds = fake_sph_flexible_grid_ds(hsml_factor=hsml_factor, nperside=3,
                                   periodic=periodic, 
                                   offsets=np.full(3, 0.5),
                                   massgenerator=makemasses,
                                   unitrho=unitrho,
                                   bbox=bbox.v, 
                                   recenter=center,
                                   e1hat=e1dir, e2hat=e2dir, e3hat=e3dir)

    source = ds.all_data()
    # couple to dataset -> right unit registry
    center = ds.arr(center, 'cm')
    print('position:\n', source['gas','position'])

    # m / rho, factor 1. / hsml**2 is included in the kernel integral
    # (density is adjusted, so same for center particle)
    prefactor = 1. / unitrho #/ (0.5 * 0.5)**2
    dl_cen = _integrate_kernel(_cubicspline_python, 0., 0.25)
    # # small rotation around y axis: auto north should be the y axis
    # # 'wrapped' particle center won't quite match the projection l.o.s.
    # # offsetx = Delta e1-coord - Delta x
    # offsetx_pwrap_nonorth = 1. / np.sum(e2dir * np.array([1., 0., 0])) - 1.
    # offsety_pwrap_nonorth = 0.
    # offsetxy_pwrap_nonorth = np.sqrt(offsetx_pwrap_nonorth**2 
    #                                  + offsety_pwrap_nonorth**2)
    # dl_pwrapx_nonorth = _integrate_kernel(_cubicspline_python, 
    #                                       offsetx_pwrap_nonorth, 0.25)
    # dl_pwrapy_nonorth = _integrate_kernel(_cubicspline_python, 
    #                                       offsety_pwrap_nonorth, 0.25)
    # dl_pwrapxy_nonorth = _integrate_kernel(_cubicspline_python, 
    #                                        offsetxy_pwrap_nonorth, 0.25)
    
    # offsetx_pwrap_north = 1. / np.sum(e2dir * np.array([1., 0., 0.])) - 1.
    # offsety_pwrap_north = 1. / np.sum(e3dir * np.array([0., 1., 0.])) - 1.
    # offsetxy_pwrap_north = np.sqrt(offsetx_pwrap_north**2 
    #                                + offsety_pwrap_north**2)
    # dl_pwrapx_north = _integrate_kernel(_cubicspline_python, 
    #                                     offsetx_pwrap_north, 0.25)
    # dl_pwrapy_north = _integrate_kernel(_cubicspline_python, 
    #                                     offsety_pwrap_north, 0.25)
    # dl_pwrapxy_north = _integrate_kernel(_cubicspline_python, 
    #                                      offsetxy_pwrap_north, 0.25)

    # we don't actually want a plot, it's just a straightforward, 
    # common way to get an frb / image array
    if weighted: 
        toweight_field = ("gas", "density")
    else:
        toweight_field = None
    prj = yt.ProjectionPlot(ds, projaxis, ("gas", "density"),
                            width=(2.5, "cm"),
                            weight_field=toweight_field,
                            buff_size=(5, 5),
                            center=center,
                            data_source=source,
                            north_vector=northvector,
                            depth=depth)
    img = prj.frb.data[('gas', 'density')]
    if weighted:
        # periodic shifts will modify the (relative) dl values a bit
        expected_out = np.zeros((5, 5,), dtype=img.v.dtype)
        expected_out[::2, ::2] = unitrho
        if depth is None: 
            ## during shift, particle coords do wrap around edges
            #if (not periodic) and shiftcenter:
            #    # weight 1. for unitrho, 2. for 2. * untrho
            #    expected_out[2, 2] *= 5. / 3.
            #else:
            # weight (2 * 1.) for unitrho, (1 * 2.) for 2. * unitrho
            expected_out[2, 2] *= 1.5
        else:
            # only 2 * unitrho element included
            expected_out[2, 2] *= 2.
    else:
        expected_out = np.zeros((5, 5,), dtype=img.v.dtype)
        expected_out[::2, ::2] = dl_cen * prefactor * unitrho
        if depth is None:
            ## during shift, particle coords do wrap around edges
            #if (not periodic) and shiftcenter:
            #    # 2 particles per l.o.s., including the denser one
            #    expected_out *= 2.
            #    expected_out[2, 2] *= 1.5
            #else:
            # 3 particles per l.o.s., including the denser one
            expected_out *= 3.
            expected_out[2, 2] *= 4. / 3.
        else:
            # 1 particle per l.o.s., including the denser one
            expected_out[2, 2] *= 2.
    # grid is shifted to the left -> 'missing' stuff at the left
    if (not periodic) and shiftcenter: 
        expected_out[:1, :] = 0.
        expected_out[:, :1] = 0.
    #print(axis, shiftcenter, depth, periodic, weighted)
    print('expected:\n', expected_out)
    print('recovered:\n', img.v)
    assert_rel_equal(expected_out, img.v, 4)

# looks good!
def visual_check_offaxis_projection():
    '''
    Plot projections along the x, z axes using the axis-aligned
    methods, then check against just-off x, z and intermediate
    lines of sight to see if we get a nice, continuous rotation.
    Note that paths here are system-specific.
    '''
    simpath = '/Users/nastasha/code/ytdev_misc/sample_data/gizmo_mhd_mwdisk/'
    simfile = 'gizmo_mhd_mwdisk.hdf5'
    outdir = '/Users/nastasha/code/ytdev_misc/testimgs/'
    outfilen_temp = 'sph_projection_test_lineofsight_{projdir}.pdf'

    nsteps = 9
    projdirs = ['x', 'z'] + [f'step{i}' for i in range(nsteps)]
    epsilon = 1e-4
    north_vector = [0., 1., 0.] #y-axis
    
    ds = yt.load(simpath + simfile)
    source = ds.all_data()
    center = [0., 0., 0.]

    for projdir in projdirs:
        if projdir in ['x', 'y', 'z']:
            outfilen = outdir + outfilen_temp.format(projdir=projdir)
            projaxis = projdir
            title = f'line of sight along the {projdir}-axis'
            #dle = ds.domain_left_edge.to("kpc")
            center = ds.arr(center, 'kpc')
            le = center - ds.quan(30., 'kpc')
            re = center + ds.quan(30, 'kpc')
            reg = YTRegion(center.to('code_length'), le.to('code_length'),
                           re.to('code_length'), ds=ds)
            # need to specify width *and* region because some giant SPH
            # particle intersects the central region
            prj = yt.ProjectionPlot(ds, projaxis, ("gas", "density"),
                                    weight_field=None,
                                    buff_size=(400, 400),
                                    center=center,
                                    width=(60, "kpc"),
                                    data_source=reg)
        else:
            _projdir = projdir + f'of{nsteps}'
            outfilen = outdir + outfilen_temp.format(projdir=_projdir)
            stepnum = int(projdir[4:])
            angle = stepnum / (nsteps - 1.) * 0.5 * np.pi
            projaxis = [np.cos(angle + epsilon), 0., np.sin(angle + epsilon)]
            title = f'line of sight x- to z-axis: step {stepnum} / {nsteps}'
            prj = yt.ProjectionPlot(ds, projaxis, ("gas", "density"),
                                    width=(60, "kpc"),
                                    weight_field=None,
                                    buff_size=(400, 400),
                                    center=center,
                                    data_source=source,
                                    north_vector=north_vector,
                                    depth=(60, 'kpc'))
        
        prj.annotate_title(title)
        prj.save(outfilen)